# id3fs - a FUSE-based filesystem for browsing audio metadata
# Copyright (C) 2010  Ian Beckwith <ianb@erislabs.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

package ID3FS::Path::Node;

use strict;
use warnings;

sub new
{
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self={};
    bless($self,$class);

    my $db=shift;
    $self->{type}=shift;
    $self->{name}=shift;
    $self->{parents_id}=shift;
    if($self->{type} ne "boolean")
    {
	my $table=ucfirst($self->{type});
	$table .= "s" unless($table=~/s$/);
	$self->{id}=$db->lookup_id($table, $self->{name}, $self->{parents_id});
	return undef unless(defined($self->{id}));
    }
    return $self;
}

sub set
{
    my($self, $name, $val)=@_;
    if(defined($val))
    {
	$self->{$name}=$val;
    }
    return $self->{$name};
}

sub left       { return shift->set("left",       shift); }
sub right      { return shift->set("right",      shift); }
sub name       { return shift->set("name",       shift); }
sub type       { return shift->set("type",       shift); }
sub id         { return shift->set("id",         shift); }
sub parents_id { return shift->set("parents_id", shift); }

sub print
{
    my($self)=@_;
    my $op=$self->name();
    my $left=$self->left();
    my $right=$self->right();
    return "" unless($left || $right);
    my $str .= $self->print_node($left);
    $str .= (" " . $op . " ") if($op);
    $str .= $self->print_node($right);
    if($op || ($left && $right))
    {
	$str="(" . $str . ")";
    }
    return $str;
}

sub print_node
{
    my($self, $node)=@_;
    return "" unless(defined($node));
    return $node->print() if(ref($node) eq "ID3FS::Path::Node");
    return $node->{name};
}

sub to_sql
{
    my($self, $hasvals, $not, @joins)=@_;
    $not=0 unless(defined($not));
    my @outjoins=();
    my $str='';
    # init
    unless(@joins)
    {
	@outjoins = @joins = ("INNER");
    }

    if($self->type() ne "boolean")
    {
	my $cnt=scalar(@joins)+1;
	$str .= "t" . scalar(@joins) . ".id='" . $self->{id} . "'";
	if($not && !$hasvals)
	{
	    $str .= " AND fxt" . scalar(@joins) . ".files_id IS NULL";
	}
	return ($str, @outjoins);
    }

    my $left=$self->left();
    my $right=$self->right();
    return ("", @outjoins) unless($left || $right);
    my ($leftstr, @leftjoins) = $left->to_sql($hasvals, $not, @joins) if($left);
    push(@joins, @leftjoins);
    push(@outjoins, @leftjoins);
    my $op=$self->name();
    print "op: $op type: ", $self->type(), " not: $not\n";
    if(defined($op))
    {
	# if we are ANDing, add an inner join
	# also if we are NOTing, but we are looking for a tag *value*
	if($op eq "AND")
	{
	    print "AND\n";
	    # hack - if right child is a NOT, we don't need extra join/brackets
	    # NOT will do the same and we will end up with an extra one
	    unless($right && $right->name() && $right->name() eq "NOT")
	    {
		push(@joins, "INNER");
		push(@outjoins, "INNER");
	    }
	}
	elsif($op eq "NOT")
	{
	    print "NOT (was $not)\n";
	    $not=1;
	    # as above - if right child is a NOT, we don't need extra join/brackets
	    # NOT will do the same and we will end up with an extra one
	    unless($right && $right->name() && $right->name() eq "NOT")
	    {
		if($hasvals)
		{
		    push(@joins, "INNER");
		    push(@outjoins, "INNER");
		}
		else
		{
		    push(@joins, "LEFT");
		    push(@outjoins, "LEFT");
		}
	    }
#	    print("LEFT: ", $left->print(), "\n") if ($left);
#	    print("RIGHT: ", $right->print(), "\n") if($right);
	}
	elsif($op eq "OR")
	{
	    print "OR\n";
	    # if left child is a NOT, we need an extra (inner) join
	    # unless right child is also a NOT
	    if(($left && $left->name() && $left->name() eq "NOT") &&
	       !($right && $right->name() && $right->name() eq "NOT"))
	    {
		push(@joins, "INNER");
		push(@outjoins, "INNER");
	    }
	}
    }
    my ($rightstr, @rightjoins) = $right->to_sql($hasvals, $not, @joins) if($right);
    push(@outjoins, @rightjoins);
#    print "LEFT (", scalar(@leftjoins), "): $leftstr\n";
#    print "RIGHT (", scalar(@rightjoins), "): $rightstr\n";
    $str=$leftstr;
    $str .= " $op " if($op && !$not);
    $str .= $rightstr;
    if($op || ($left && $right))
    {
	$str="(" . $str . ")";
    }
#    print "STR: $str\n";
#    my @all=(@joins, @rightjoins);
#    print "JOINS: RETURN ", scalar(@outjoins), " ALL ", scalar(@all), "\n";
    return($str, @outjoins);
}

sub used_tags
{
    my($self)=@_;
    if($self->type() eq "boolean")
    {
	my @used=();
	push(@used, $self->left()->used_tags())  if($self->left());
	push(@used, $self->right()->used_tags()) if($self->right());
	return(grep { defined; } @used);
    }
    return $self->id();
}

1;
